# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

_PLACEHOLDERS = {
    '{TITLE}'     : lambda e: e.title,
    '{USERNAME}'  : lambda e: e.username,
    '{URL}'       : lambda e: e.url,
    '{PASSWORD}'  : lambda e: e.password,
    '{NOTES}'     : lambda e: e.notes,
    '{PLUS}'      : lambda _: '+',
    '{PERCENT}'   : lambda _: '%',
    '{CARET}'     : lambda _: '^',
    '{TILDE}'     : lambda _: '~',
    '{LEFTPAREN}' : lambda _: '(',
    '{RIGHTPAREN}': lambda _: ')',
    '{LEFTBRACE}' : lambda _: '{',
    '{RIGHTBRACE}': lambda _: '}',
    '{AT}'        : lambda _: '@',
    '{+}'         : lambda _: '+',
    '{%}'         : lambda _: '%',
    '{^}'         : lambda _: '^',
    '{~}'         : lambda _: '~',
    '{(}'         : lambda _: '(',
    '{)}'         : lambda _: ')',
    '{[}'         : lambda _: '[',
    '{]}'         : lambda _: ']',
    '{{}'         : lambda _: '{',
    '{}}'         : lambda _: '}',
}

def replace_placeholder(entry, token):
    fn = _PLACEHOLDERS.get(token)
    if callable(fn):
        return fn(entry)
    return None




# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import os
import sys
import subprocess
import operator
import collections


class DatabaseError(Exception):
    pass

# Delays importing of PyKeePass which by itself adds ~150 ms to startup...
class DelayedPyKeePass:
    def __init__(self, args, ioh):
        self._db = None
        if args.db:
            self.change_db(args.db)

        self._password = args.password
        self._keyfile = args.keyfile

        self._pw_cmd = args.pw_cmd
        self._pinentry = args.pinentry

        self._kp = None
        self._mtime = None
        self._entries  = None

        self._ioh = ioh

    def _getpass(self):
        if self._password is not None:
            yield self._password
        if self._pw_cmd is not None:
            yield readpass(self._pw_cmd)
        if self._password is None and self._pw_cmd is None \
                and self._keyfile is not None:
            # password=None supports Composite Master Key with only a keyfile
            # https://keepass.info/help/base/keys.html
            yield None
        yield askpass(self._db, self._pinentry, self._ioh)

    def unlock(self):
        if self._kp and self._mtime == self._read_mtime():
            return True
        elif self._kp:
            # If e.g. database password was changed, this might fail. That's OK, 
            # but we must allow user to provide a password in a second,
            # "ordinary" database read.
            if self._read_db([self._kp.password], self._kp.keyfile):
                return True
            else:
                self._ioh.eprint(
                    'Database was modified externally and must be reloaded, '
                    'but auto-reload failed. Re-obtaining password.')

        if not self._db:
            self._ioh.eprint('No database is selected.')
            return False

        if not self._read_db(self._getpass, self._keyfile):
            self._ioh.eprint('Failed unlocking database {}.'.format(self._db))
            return False
        return True

    def lock(self):
        self._kp = None
        self._entries = None
        self._password = None
        self._keyfile = None

    def change_db(self, db):
        self.lock()
        if not os.path.isfile(db):
            self._ioh.eprint('Invalid database: \'{}\''.format(db))
            return
        self._db = db

    def change_credentials(self, password=None, keyfile=None):
        if not self.locked:
            raise DatabaseError(
                'Database must be locked before changing credentials '
                'used for unlocking.')

        self._password = password
        self._keyfile = keyfile

    @property
    def locked(self):
        return self._kp is None

    @property
    def db(self):
        return self._db

    @property
    def kp(self):
        self.unlock()
        return self._kp

    @property
    def entries(self):
        if not self.kp:
            return collections.OrderedDict()

        if self._entries is None:
            self._entries = collections.OrderedDict(
                (e.path, e)
                for e in sorted(self.kp.entries,
                                key=operator.attrgetter('path')))

        return self._entries

    @property
    def paths(self):
        return [e for e in self.entries.keys()]

    def iter_paths(self):
        for entry in self.entries:
            yield entry

    def _read_mtime(self):
        return os.stat(self._db).st_mtime

    def _read_db(self, password_iter, keyfile):
        from pykeepass import PyKeePass

        if callable(password_iter):
            password_iter = password_iter()

        kp = None
        for pw in password_iter:
            try:
                kp = PyKeePass(self._db, password=pw, keyfile=keyfile)
            except Exception as e:
                pass
            else:
                self._mtime = self._read_mtime()
                break

        self._kp = kp
        self._entries = None
        return kp is not None


def askpass(db, pinentry, ioh):
    if pinentry:
        return _askpass_pinentry(db, pinentry)
    return _askpass_prompt(ioh)


def readpass(cmd):
    cp = subprocess.run(cmd, capture_output=True, text=True, shell=True)
    if cp.returncode != 0 or not cp.stdout:
        return None
    return cp.stdout


def _askpass_pinentry(db, pinentry):
    pein = 'setdesc Enter password for {}\ngetpin\n'.format(db)
    cp = subprocess.run(pinentry, input=pein, capture_output=True,
                        text=True)

    if cp.returncode != 0 or not cp.stdout:
        return None

    lines = cp.stdout.splitlines()
    dlines = [line for line in lines if line.startswith ('D ')]
    if not dlines:
        return None

    passline = dlines[0]
    return passline.partition(' ')[-1]


def _askpass_prompt(ioh):
    return ioh.prompt('Password: ', is_password=True)



import os
import shutil
from unittest.mock import Mock
import pytest

from kpd.io import Handler

@pytest.fixture
def tmpdb(tmpdir):
    curdir = os.path.dirname(os.path.abspath(__file__))

    keysrc = os.path.join(curdir, 'db', 'keyfile')
    keydst = os.path.join(str(tmpdir), 'keyfile')
    shutil.copy(keysrc, keydst)

    def _f(name):
        src = os.path.join(curdir, 'db', name)
        dst = os.path.join(str(tmpdir), name)
        shutil.copy(src, dst)
        return dst
    return _f


@pytest.fixture
def fakeargs():
    class Args:
        def __getattr__(self, name):
            return self.__dict__.get(name, None)

    return Args()


@pytest.fixture
def iohandler():
    class ProggrammableIO(Handler):
        def __init__(self):
            self.prompt = Mock()
            self.get_command = Mock()
            super().__init__()

    return ProggrammableIO()



import os
import glob
import re

from prompt_toolkit.completion import Completer, Completion

_OPT_PATTERN = re.compile(r'([a-zA-Z0-9_]+|[^a-zA-Z0-9_\s]+)')

class CommandCompleter(Completer):
    def __init__(self, kp, commands):
        self._commands = sorted(commands)
        self._kp = kp

    def get_completions(self, document, complete_event):
        text_before = document.text_before_cursor
        if ' ' in text_before:
            command = text_before.partition(' ')[0]
            fn = getattr(self, '_{}'.format(command), None)
            if callable(fn):
                yield from fn(document)
        else:
            yield from self._complete_commands(document)

    def _complete_commands(self, document):
        text_before = document.text_before_cursor
        for command in self._commands:
            if command.startswith(text_before):
                yield Completion(command, -len(text_before))

    def _show(self, document):
        opts = ['--no-field-name']
        yield from self._path_and_opts(document, opts)

    def _autotype(self, document):
        opts = ['--default::change default sequence',
                '--sequence::override sequence',
                '--force::force auto-type when disabled',
                '--delay::delay between keypresses']
        yield from  self._path_and_opts(document, opts)

    def _help(self, document):
        word = document.get_word_before_cursor()
        for command in self._commands:
            if command.startswith(word):
                yield Completion(command, -len(word))

    def _open(self, document):
        yield from self._filepath(document)

    def _path_and_opts(self, document, opts):
        word = document.get_word_before_cursor(WORD=True)
        if word.startswith('-'):
            completions = sorted(opts)
        else:
            completions = self._kp.paths if not self._kp.locked else []

        for compl in completions:
            compl, _, meta = compl.partition('::')
            if not meta:
                meta = None
            if compl.startswith(word):
                yield Completion(compl, -len(word), display_meta=meta)

    def _filepath(self, document):
        word = document.get_word_before_cursor(WORD=True)

        home = os.path.expanduser('~')
        tilde_repl = word.startswith('~/')
        for compl in glob.iglob('{}*'.format(os.path.expanduser(word))):
            meta = None
            if os.path.isfile(compl):
                meta = 'file'
            elif os.path.isdir(compl):
                meta = 'dir'

            if tilde_repl:
                compl = compl.replace(home, '~')

            yield Completion(compl, -len(word), display_meta=meta)



import os
import pytest

from kpd.db import DelayedPyKeePass


@pytest.fixture
def dbargs(tmpdb, fakeargs):
    fakeargs.db = tmpdb('db_argon2_kdbx4_pass.kdbx')
    return fakeargs


def test_states(dbargs, iohandler):
    dbargs.password = 'foobar'
    kp = DelayedPyKeePass(dbargs, iohandler)
    assert kp.locked
    assert kp.unlock()
    assert not kp.locked


def test_fail_unlock(dbargs, iohandler):
    kp = DelayedPyKeePass(dbargs, iohandler)
    assert not kp.unlock()


def test_unlock_of_unlocked_db(dbargs, iohandler):
    iohandler.prompt.return_value = 'foobar'
    kp = DelayedPyKeePass(dbargs, iohandler)

    assert kp.unlock()

    iohandler.prompt.return_value = None
    assert kp.unlock()


def test_auto_reload(dbargs, iohandler):
    iohandler.prompt.return_value = 'foobar'
    kp = DelayedPyKeePass(dbargs, iohandler)

    assert kp.unlock()
    assert kp.kp

    # Let's change the passwords so we can verify that auto-reloading fails
    kp.kp.password = None
    iohandler.prompt.return_value = None

    # Modify mtime of our database
    curr = os.stat(dbargs.db)
    os.utime(dbargs.db, (curr.st_atime, curr.st_mtime + 100))

    assert not kp.kp  # failed auto-unlocking because we removed password

    iohandler.prompt.return_value = 'foobar'
    assert kp.kp



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import time
import re

def _token_delay(token):
    match = re.match(r'{DELAY (\d+)}', token)
    if match:
        delay = match.group(1)
        return lambda t=delay: time.sleep(int(t) / 1000)
    return None

_COMMANDS = [
    _token_delay,
]

def run_command(token):
    for checker in _COMMANDS:
        cmd = checker(token)
        if callable(cmd):
            cmd()
            return True
    return False




# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2019 Michał Góral.

import sys
import os
import argparse
import subprocess
import time
import fnmatch


class CommandError(Exception): pass
class ArgumentParserError(Exception): pass


class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


def prepare_command_parser():
    cp = ThrowingArgumentParser(prog='', add_help=False)
    sp = cp.add_subparsers(required=True)

    parsers = {}
    parsers[None] = cp

    # helper function which automatically creates help-friendly parsers
    def add_parser(command, *a, **kw):
        kw['add_help'] = False

        descr = kw.get('description')
        if descr:
            kw['description'] = '{}\n\n{}'.format(kw['help'], descr)
        else:
            kw['description'] = kw['help']

        parser = sp.add_parser(command, *a, **kw)
        parsers[command] = parser
        return parser

    ######### open
    open_sp = add_parser('open', help='Change currently opened database.')
    open_sp.add_argument('filepath', help='path to database file.')
    open_sp.set_defaults(func=open_)

    ######### unlock
    unlock_sp = add_parser('unlock', help='Unlock currently opened database.')
    unlock_sp.add_argument('--keyfile', default='',
                           help='key file used for unlocking database')
    unlock_sp.set_defaults(func=unlock)

    ######### lock
    lock_sp = add_parser('lock', help='Lock a database.')
    lock_sp.set_defaults(func=lock)

    ######### db
    db_sp = add_parser('db', help='Query opened database info.')
    db_sp.set_defaults(func=db)

    ######### ls
    ls_sp = add_parser('ls', help='List contents of database.')
    ls_sp.add_argument('glob', nargs='?', default='*',
                       help='display only entries which match glob expression')
    ls_sp.set_defaults(func=ls)

    ######### show
    show_sp = add_parser('show', help='Show contents of entry.',
        description='Search is case-sensitive.')
    show_sp.add_argument('path', help='path which should be shown')
    show_sp.add_argument('fields', nargs='*',
        help='only display certain fields')
    show_sp.add_argument('-n', '--no-field-name', action='store_true',
        help='hide field name when printing entry fields.')
    show_sp.set_defaults(func=show)

    ######### autotype
    at_sp = add_parser('autotype', help='Auto-type sequence of entry fields.',
        description='This simulates keypresses to any currently open window. '
                    'It\'s particularily useful when kpsh is run from a script '
                    'or keypress in non-interactive mode (`-c` switch). If '
                    '`-s` is given, it will be used as auto-type sequence. '
                    'Otherwise sequence defined for selected entry will be '
                    'used or the default one if there is none (`-d`).')
    at_sp.add_argument('path', help='path of entry to auto-type')
    at_sp.add_argument('-s', '--sequence', help='override auto-type sequence')
    at_sp.add_argument('-d', '--default',
        default='{USERNAME}{TAB}{PASSWORD}{ENTER}',
        help='default auto-type sequence used when entry doesn\'t specify '
             'sequence itself.')
    at_sp.add_argument('-D', '--delay', type=int, default=40,
        help='delay beteen simulated keypresses')
    at_sp.add_argument('-f', '--force', action='store_true',
        help='force auto-type for entries for which auto-type was disabled')
    at_sp.set_defaults(func=autotype)

    ######### exit
    exit_sp = add_parser('exit', help='Exit shell.')
    exit_sp.set_defaults(func=exit)

    ######### echo
    echo_sp = add_parser('echo', help='Display a message.')
    echo_sp.add_argument('message', nargs='*', help='message to be displayed')
    echo_sp.set_defaults(func=echo)

    ######### sleep
    sleep_sp = add_parser('sleep', help='Sleep for a given number of seconds.',
        description='Seconds might be a floating number when fractions of '
                    'second are needed.')
    sleep_sp.add_argument('secs', type=float, help='seconds to sleep')
    sleep_sp.set_defaults(func=sleep)

    ######### help
    help_sp = add_parser('help', help='Show help for any message.')
    help_sp.add_argument('command', nargs='?')
    help_sp.set_defaults(func=lambda *a, parsers=parsers: help_(*a, parsers))

    return cp, parsers

def tokenize(seq):
    tokens = []

    i = 0
    while i < len(seq):
        start = seq.find('{', i)

        if start != -1:
            if i < start:
                tokens.append(seq[i:start])

            end = seq.find('}', start)
            nend = end + 1

            if end == -1:
                end = len(seq) - 1
            elif end == start + 1 and len(seq) > nend and seq[nend] == '}':  # {}}
                end = nend

            tokens.append(seq[start:end + 1])
            i = end + 1
        else:
            tokens.append(seq[i:])
            i = len(seq)

    return tokens


def xdotool(*args):
    cmd = ['/usr/bin/xdotool']
    cmd.extend(args)
    subprocess.run(cmd)


def autotype(kp, args, ioh):
    from kpd.autotype.xdotoolkeys import XDOTOOL_KEYS
    from kpd.autotype.placeholders import replace_placeholder
    from kpd.autotype.commands import run_command

    delay = str(args.delay)
    entry = _get(args.path, kp)

    if not entry.autotype_enabled and not args.force:
        ioh.eprint('Autotype disabled for {}. '
                   'Use -f to force autotype.'.format(args.path))
        return

    sequence = args.sequence if args.sequence else entry.autotype_sequence
    if not sequence:
        sequence = args.default

    for token in tokenize(sequence):
        if token.startswith('{') and token.endswith('}'):
            if run_command(token):
                continue

            placeholder = replace_placeholder(entry, token)
            if placeholder is not None:
                xdotool('type', '--delay', delay, '--', placeholder)
                continue

            specialkey = XDOTOOL_KEYS.get(token)
            if specialkey is not None:
                xdotool('key', '--clearmodifiers', '--delay', delay, specialkey)
                continue

            ioh.eprint('Unsupported keyword: {}'.format(token))
        else:
            xdotool('type', '--delay', delay, '--', token)


def show(kp, args, ioh):
    entry = _get(args.path, kp)
    attrs = ['path', 'username', 'password', 'url', 'autotype_sequence',
              'notes']

    for attr in attrs:
        if args.fields and attr not in args.fields:
            continue
        val = getattr(entry, attr)
        if val is None:
            continue

        if args.no_field_name:
            ioh.print(val)
        else:
            ioh.print('{}: {}'.format(attr, val))


def echo(kp, args, ioh):
    ioh.print(*args.message)


def sleep(kp, args, ioh):
    time.sleep(args.secs)


def ls(kp, args, ioh):
    for path in fnmatch.filter(kp.iter_paths(), args.glob):
        ioh.print(path)


def help_(kp, args, ioh, parsers):
    parser = parsers.get(args.command)
    if parser is None:
        ioh.eprint('no such command: {}'.format(args.command))
        parser = parsers.get(None)

    parser.print_help()


def exit(kp, args, ioh):
    ioh.stop()


def open_(kp, args, ioh):
    fp = os.path.expanduser(args.filepath)
    kp.change_db(fp)


def unlock(kp, args, ioh):
    if not kp.locked:
        return

    kf = os.path.expanduser(args.keyfile) if args.keyfile else None
    kp.change_credentials(keyfile=kf)
    kp.unlock()


def lock(kp, args, ioh):
    kp.lock()


def db(kp, args, ioh):
    ioh.print(kp.db)
    ioh.print('Locked: {}'.format(kp.locked))


def _get(path, kp):
    entry = kp.entries.get(path)
    if not entry:
        raise CommandError('entry not found: {}'.format(path))
    return entry



import os
import sys
import shlex
from functools import lru_cache
from contextlib import contextmanager
import socket
import struct
import select

from prompt_toolkit import PromptSession
from prompt_toolkit import prompt as input_prompt
from prompt_toolkit.formatted_text.html import HTML

from kpd.commands import CommandError, ArgumentParserError
from kpd.completion import CommandCompleter

class Handler:
    def __init__(self):
        self._run = True
        self._returncode = 0

    def prompt(self, prompt='', is_password=False):
        '''Used to prompt for any kind of input'''
        return None

    def print(self, *a, **kw):
        '''Used to print ordinary messages'''
        print(*a, **kw)

    def eprint(self, *a, **kw):
        '''Used to print error messages'''
        kw['file'] = sys.stderr
        print(*a, **kw)

    def get_command(self, kp):
        '''Called when handler is ready to accept a next command. Returns a
        unparsed argument-like string. Children should override this method to
        read user input.'''
        return None

    def finalize_command(self):
        '''Always called after each command ends.'''
        pass

    def initialize(self, kp):
        '''Called before Handler enters its command loop.'''
        pass

    def teardown(self, kp):
        '''Called after Handler exits its command loop.'''
        pass

    def stop(self, returncode=0):
        '''Used to stop the running command loop.'''
        self._run = False
        self._returncode = returncode

    def run(self, kp, cmd_parser):
        with self._setup(kp):
            while self._run:
                text = self.get_command(kp)
                if not text:
                    continue

                cmd = shlex.split(text)
                if not cmd:
                    continue

                try:
                    cargs = cmd_parser.parse_args(cmd)
                    cargs.func(kp, cargs, self)
                except ArgumentParserError as e:
                    self.eprint(str(e))
                except CommandError as e:
                    self.eprint(str(e))
                finally:
                    self.finalize_command()
        return self._returncode

    @contextmanager
    def _setup(self, kp):
        self.initialize(kp)
        try:
            yield
        except KeyboardInterrupt:
            self.stop(1)
        finally:
            self.teardown(kp)


class Interactive(Handler):
    def __init__(self, prompt, parsers):
        self._session = None
        self._prompt = prompt
        self._parsers = parsers
        super().__init__()

    def initialize(self, kp):
        compl = CommandCompleter(kp, (p for p in self._parsers if p))
        self._session = PromptSession(completer=compl,
                                      complete_while_typing=False)

    def get_command(self, kp):
        try:
            return self._session.prompt(self._ps1(kp.db))
        except KeyboardInterrupt:
            return None
        except EOFError:
            self.stop()
            return None

    def prompt(self, prompt='', is_password=False):
        try:
            return input_prompt(prompt, is_password=is_password)
        except (EOFError, KeyboardInterrupt):
            return None

    @lru_cache(maxsize=32)
    def _ps1(self, dbpath):
        if not dbpath:
            dbpath = ''
        home = os.path.expanduser('~')
        if dbpath.startswith(home):
            dbpath = dbpath.replace(home, '~')

        return HTML(self._prompt.format(dbpath))


class NonInteractive(Handler):
    def __init__(self):
        self._read_input = False
        super().__init__()

    def get_command(self, kp):
        try:
            return input()
        except EOFError:
            self.stop()
            return None


class Arguments(Handler):
    def __init__(self, commands):
        self._iter = iter(commands)
        super().__init__()

    def get_command(self, kp):
        try:
            return next(self._iter)
        except StopIteration:
            self.stop()
            return None


class Daemon(Handler):
    def __init__(self, path):
        self._path = path
        super().__init__()

    def initialize(self, kp):
        self._serv = _SocketServer(self._path)

    def teardown(self, kp):
        self._serv.stop()

    def prompt(self, prompt='', is_password=False):
        if is_password:
            resptype = 'PS'
        else:
            resptype = 'P'

        msg = '{} {}'.format(resptype, prompt)
        self._send(msg)
        return self._recv()

    def print(self, *a, **kw):
        msg = 'M {}'.format(' '.join(str(arg) for arg in a))
        self._send(msg)

    def eprint(self, *a, **kw):
        msg = 'E {}'.format(' '.join(a))
        self._send(msg)

    def get_command(self, kp):
        return self._recv()

    def finalize_command(self):
        self._send('OK')

    def _recv(self):
        buf = self._serv.recv(4)
        if not buf:
            return None

        msglen = struct.unpack("!i", buf)[0]
        if msglen == 0:
            return None

        data = self._serv.recv(msglen)
        if not data:
            return None

        return data.decode()

    def _send(self, msg):
        data = msg.encode()
        msglen = struct.pack('!i', len(data))
        self._serv.send(msglen)
        self._serv.send(data)


class _SocketServer:
    '''Wrapper for all hairy socket stuff'''

    def __init__(self, path):
        self._remove_socket(path)
        self._sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._sock.bind(path)
        self._sock.listen(1)
        self._conn = None

    def send(self, msg):
        if not self._conn:
            return

        msglen = len(msg)
        totalsent = 0
        while totalsent < msglen:
            try:
                sent = self._conn.send(msg[totalsent:msglen])
            except BrokenPipeError:
                sent = 0

            if sent == 0:
                self._close_conn()
                return
            totalsent += sent

    def recv(self, size):
        self._wait_for_client()
        buf = self._conn.recv(size)
        if not buf:
            self._close_conn()
            return None
        return buf

    def stop(self):
        self._close_conn()

        if self._sock:
            self._sock.shutdown(socket.SHUT_RDWR)
            self._sock.close()
            self._sock = None

    def _accept(self):
        conn, _ = self._sock.accept()
        return conn

    def _wait_for_client(self):
        if not self._conn:
            self._conn = self._accept()

        while True:
            # Server continuously accepts all connections because queuing them
            # on listen() is unreliable - it's impossible to tell how listen(0)
            # behaves. Certainly it queues at least 1 non-accepted connection,
            # which might block clients forever. So we'll just wait on
            # select() whenever blocking socket operation occurs and close all
            # connections but the first one. Clients should notice broken
            # connection with first recv().
            readable, _, _ = select.select([self._conn, self._sock], [], [])
            if self._sock in readable:
                self._accept().close()
            if self._conn in readable:
                break

    def _close_conn(self):
        if not self._conn:
            return

        self._conn.close()
        self._conn = None

    def _remove_socket(self, path):
        try:
            os.remove(path)
        except FileNotFoundError:
            pass



# SPDX-License-Identifier: GPL-3.0-or-later

# Copyright (C) 2019 Michał Góral.

import os
import sys
import argparse
import tempfile
import signal

from kpd.io import Interactive, NonInteractive, Daemon, Arguments

from kpd.db import DelayedPyKeePass
from kpd.commands import prepare_command_parser
from kpd._version import version


class Cleanup:
    def __init__(self):
        signal.signal(signal.SIGINT, self.sysexit)
        signal.signal(signal.SIGTERM, self.sysexit)

    def sysexit(self, signum, frame):
        raise SystemExit(signum)


def default_socket_path():
    tmp = tempfile.gettempdir()
    fname = 'kpsh-{}.sock'.format(os.getuid())
    return os.path.join(tmp, fname)


def prepare_args():
    ap = argparse.ArgumentParser(description='KeePass database shell access.')
    ap.add_argument('db', nargs='?', help='path to KeePass database.')

    ap.add_argument('--password', default=None,
        help='Database password.')
    ap.add_argument('--pw-cmd', default=None,
        help='Password will be obtained from the output of this command.')
    ap.add_argument('--keyfile', default=None,
        help='Key file for unlocking database.')
    ap.add_argument('--pinentry',
        help='Command used to run pinentry.')
    ap.add_argument('-c', '--command', action='append',
        help='Command to execute. If command contains spaces, it must be '
             'enclosed in double quotes. kpsh will be started in '
             'non-interactive mode.')
    ap.add_argument('--prompt', default='<style fg="ansiblue">{}</style>> ',
        help='Text used by shell for prompt.')
    ap.add_argument('-d', '--daemon', action='store_true',
        help='Start as daemon listening on a socket given by --socket-path')
    ap.add_argument('-s', '--socket-path', default=default_socket_path(),
        help='Path to the socket which will be created in daemon mode.')
    ap.add_argument('--version', action='version',
                    version='%(prog)s {}'.format(version))
    return ap.parse_args()


def main():
    cl = Cleanup()
    args = prepare_args()

    cmd_parser, parsers = prepare_command_parser()

    if args.command:
        ioh = Arguments(args.command)
    elif args.daemon:
        ioh = Daemon(args.socket_path)
    elif sys.stdin.isatty():
        ioh = Interactive(args.prompt, parsers)
    else:
        ioh = NonInteractive()

    kp = DelayedPyKeePass(args, ioh)
    return ioh.run(kp, cmd_parser)
